#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#          Calculate : Volumetric Similarity          #
#-----------------------------------------------------#
"""
Formula:
    VS = 1 - (|FN-FP| / (2TP + FP + FN))

References:
    Taha, A.A., Hanbury, A.
    Metrics for evaluating 3D medical image segmentation: analysis, selection, and tool.
    BMC Med Imaging 15, 29 (2015). https://doi.org/10.1186/s12880-015-0068-x
"""
def calc_VolumetricSimilarity(truth, pred, c=1):
    # Obtain confusion mat
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c)
    # Compute VS
    if (2*tp + fp + fn) != 0:
        vs = 1 - (np.abs(fn-fp) / (2*tp + fp + fn))
    else : vs = 1.0 - 0.0
    # Return VS score
    return vs



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Metric Imports                    #
#-----------------------------------------------------#
# Confusion Matrix
from miseval.confusion_matrix import *
# Dice Similarity Coefficient
from miseval.dice import *
from miseval.dice import calc_DSC_Sets as calc_DSC
# Intersection-over-Union
from miseval.jaccard import *
from miseval.jaccard import calc_IoU_Sets as calc_IoU
# Sensitivity
from miseval.sensitivity import *
from miseval.sensitivity import calc_Sensitivity_Sets as calc_Sensitivity
# Specificity
from miseval.specificity import *
from miseval.specificity import calc_Specificity_Sets as calc_Specificity
# Precision
from miseval.precision import *
from miseval.precision import calc_Precision_Sets as calc_Precision
# Accuracy
from miseval.accuracy import *
from miseval.accuracy import calc_Accuracy_Sets as calc_Accuracy
# Area under the ROC
from miseval.auc import *
from miseval.auc import calc_AUC_trapezoid as calc_AUC
# Cohen's Kappa
from miseval.kappa import *
# Hausdorff Distance
from miseval.hausdorff import calc_SimpleHausdorffDistance, \
                              calc_AverageHausdorffDistance
# Volumetric Similarity
from miseval.volumetric_similarity import *

#-----------------------------------------------------#
#         Access Functions to Metric Functions        #
#-----------------------------------------------------#
# Metric Dictionary
metric_dict = {
    "TruePositive": calc_TruePositive,
    "TrueNegative": calc_TrueNegative,
    "FalsePositive": calc_FalsePositive,
    "FalseNegative": calc_FalseNegative,
    "TP": calc_TruePositive,
    "TN": calc_TrueNegative,
    "FP": calc_FalsePositive,
    "FN": calc_FalseNegative,
    "DSC": calc_DSC,
    "Dice": calc_DSC,
    "DiceSimilarityCoefficient": calc_DSC,
    "IoU": calc_IoU,
    "Jaccard": calc_IoU,
    "IntersectionOverUnion": calc_IoU,
    "ACC": calc_Accuracy,
    "Accuracy": calc_Accuracy,
    "RI": calc_Accuracy,
    "RandIndex": calc_Accuracy,
    "IntersectionOverUnion": calc_IoU,
    "IntersectionOverUnion": calc_IoU,
    "BACC": calc_BalancedAccuracy,
    "BalancedAccuracy": calc_BalancedAccuracy,
    "ARI": calc_AdjustedRandIndex,
    "AdjustedRandIndex": calc_AdjustedRandIndex,
    "AUC": calc_AUC,
    "AUC_trapezoid": calc_AUC,
    "KAP": calc_Kappa,
    "Kappa": calc_Kappa,
    "CohensKappa": calc_Kappa,
    "Sensitivity": calc_Sensitivity,
    "SENS": calc_Sensitivity,
    "TPR": calc_Sensitivity,
    "TruePositiveRate": calc_Sensitivity,
    "Recall": calc_Sensitivity,
    "SPEC": calc_Specificity,
    "Specificity": calc_Specificity,
    "TNR": calc_Specificity,
    "TrueNegativeRate": calc_Specificity,
    "PREC": calc_Precision,
    "Precision": calc_Precision,
    "VS": calc_VolumetricSimilarity,
    "VolumetricSimilarity": calc_VolumetricSimilarity,
    "HD": calc_SimpleHausdorffDistance,
    "HausdorffDistance": calc_SimpleHausdorffDistance,
    "AHD": calc_AverageHausdorffDistance,
    "AverageHausdorffDistance": calc_AverageHausdorffDistance
}

#-----------------------------------------------------#
#                     Core Imports                    #
#-----------------------------------------------------#
from miseval.core import evaluate



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
import unittest
# Internal modules
from miseval import evaluate, metric_dict

#-----------------------------------------------------#
#             Unittest: Core Evaluate Hub             #
#-----------------------------------------------------#
class TEST_CoreEvaluate(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Create ground truth
        np.random.seed(1)
        self.gt_bi = np.random.randint(2, size=(32,32))
        self.gt_mc = np.random.randint(5, size=(32,32))
        # Create prediction
        np.random.seed(2)
        self.pd_bi = np.random.randint(2, size=(32,32))
        self.pd_mc = np.random.randint(5, size=(32,32))

    #-------------------------------------------------#
    #         Evaluate : Metric Functionality         #
    #-------------------------------------------------#
    def test_evaluate_metric(self):
        # Check miseval function callings
        for metric in metric_dict:
            self.assertTrue(metric in metric_dict)
            self.assertTrue(callable(metric_dict[metric]))
            evaluate(self.gt_bi, self.pd_bi, metric)
            evaluate(self.gt_bi, self.pd_bi, metric_dict[metric])
        # Check custom function
        def my_custom_function(g,p,c) : return 0.0
        evaluate(self.gt_bi, self.pd_bi, my_custom_function)
        # Check lower string metric calling
        evaluate(self.gt_bi, self.pd_bi, "tp")
        evaluate(self.gt_bi, self.pd_bi, "fn")
        # Provide some non-existent metric string
        self.assertRaises(KeyError, evaluate, self.gt_bi, self.pd_bi, "test")
        # Provide a non callable metric
        self.assertRaises(ValueError, evaluate, self.gt_bi, self.pd_bi, 7)

    #-------------------------------------------------#
    #          Evaluate : Binary with Binary          #
    #-------------------------------------------------#
    def test_evaluate_binary_bi(self):
        for metric in metric_dict:
            score = evaluate(self.gt_bi, self.pd_bi, metric,
                             multi_class=False, n_classes=2)
            self.assertTrue(isinstance(score, np.int64) or \
                            isinstance(score, np.float64))

    #-------------------------------------------------#
    #        Evaluate : Binary with Multi-Class       #
    #-------------------------------------------------#
    def test_evaluate_binary_mc(self):
        for metric in metric_dict:
            scores = evaluate(self.gt_mc, self.pd_mc, metric,
                              multi_class=False, n_classes=5)
            self.assertTrue(len(scores) == 5)
            self.assertTrue(isinstance(scores, np.ndarray))

    #-------------------------------------------------#
    #        Evaluate : Multi-Class with Binary       #
    #-------------------------------------------------#
    def test_evaluate_multiclass_bi(self):
        for metric in metric_dict:
            scores = evaluate(self.gt_bi, self.pd_bi, metric,
                              multi_class=True, n_classes=2)
            self.assertTrue(len(scores) == 2)
            self.assertTrue(isinstance(scores, np.ndarray))

    #-------------------------------------------------#
    #     Evaluate : Multi-Class with Multi-Class     #
    #-------------------------------------------------#
    def test_evaluate_multiclass_mc(self):
        for metric in metric_dict:
            scores = evaluate(self.gt_mc, self.pd_mc, metric,
                              multi_class=True, n_classes=5)
            self.assertTrue(len(scores) == 5)
            self.assertTrue(isinstance(scores, np.ndarray))

    #-------------------------------------------------#
    #           Evaluate : Exception Testing          #
    #-------------------------------------------------#
    def test_evaluate_exceptions(self):
        # Check for incorrect number of classes in ground truth
        with self.assertRaises(ValueError):
            scores = evaluate(self.gt_mc, self.pd_bi, "TruePositive",
                              multi_class=True, n_classes=2)
        # Check for incorrect number of classes in prediction
        with self.assertRaises(ValueError):
            scores = evaluate(self.gt_bi, self.pd_mc, "TruePositive",
                              multi_class=True, n_classes=2)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External modules
import numpy as np
# Internal modules
from miseval.confusion_matrix import calc_ConfusionMatrix

#-----------------------------------------------------#
#           Calculate : Sensitivity via Sets          #
#-----------------------------------------------------#
def calc_Sensitivity_Sets(truth, pred, c=1):
    # Obtain sets with associated class
    gt = np.equal(truth, c)
    pd = np.equal(pred, c)
    # Calculate sensitivity
    if gt.sum() != 0 : sens = np.logical_and(pd, gt).sum() / gt.sum()
    else : sens = 0.0
    # Return sensitivity
    return sens

#-----------------------------------------------------#
#            Calculate : Sensitivity via CM           #
#-----------------------------------------------------#
def calc_Sensitivity_CM(truth, pred, c=1):
    # Obtain confusion matrix
    tp, tn, fp, fn = calc_ConfusionMatrix(truth, pred, c)
    # Calculate sensitivity
    if (tp + fn) != 0 : sens = (tp) / (tp + fn)
    else : sens = 0.0
    # Return sensitivity
    return sens
